# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import List, Optional

import dataclasses_json
from card_framework import Renderable, list_field, standard_field

from .widget import Widget


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class Section(Renderable):
  """Section

  Describes a Google Chat App response Section.

  https://developers.google.com/chat/api/guides/message-formats/cards#sections_and_widgets
  """
  __SUPPRESS_TAG__ = True

  header: Optional[str] = standard_field()
  widgets: Optional[List[Widget]] = list_field(default_factory=list)
  collapsible: bool = standard_field()
  uncollapsible_widgets_count: int = standard_field()

  def add_widget(self, widget: Widget) -> None:
    """Adds a widget to the section.

    A helper to replicate section.widgets.append(<widget>) as this can be
    more readable and apparent to an end user.

    Args:
        widget (Widget): The widget to be added to the section.
    """
    self.widgets.append(widget)



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses

import dataclasses_json
from card_framework import AutoNumber, enum_field, standard_field


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class ActionStatus(object):
  """ActionStatus
  """
  class Code(AutoNumber):
    OK = ()
    CANCELLED = ()
    UNKNOWN = ()
    INVALID_ARGUMENT = ()
    DEADLINE_EXCEEDED = ()
    NOT_FOUND = ()
    ALREADY_EXISTS = ()
    PERMISSION_DENIED = ()
    UNAUTHENTICATED = ()
    RESOURCE_EXHAUSTED = ()
    FAILED_PRECONDITION = ()
    ABORTED = ()
    OUT_OF_RANGE = ()
    UNIMPLEMENTED = ()
    INTERNAL = ()
    UNAVAILABLE = ()
    DATA_LOSS = ()

  status_code: Code = enum_field()
  user_facing_message: str = standard_field()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import enum
import inspect
from typing import Any, Callable, Mapping

import dataclasses_json
import stringcase


def lazy_property(f: Callable):
  """Decorator that makes a property lazy-evaluated.

  Args:
    f: the function to convert to a lazy property.
  """
  attr_name = '_lazy_' + f.__name__

  @property
  def _lazy_property(self) -> Any:
    if not hasattr(self, attr_name):
      setattr(self, attr_name, f(self))
    return getattr(self, attr_name)
  return _lazy_property


def field(default: Any = None, default_factory: Any = None,
          **metadata) -> dataclasses.Field:
  return (
      dataclasses.field(
          default_factory=default_factory,
          metadata=dataclasses_json.config(**metadata)
      ) if default_factory
      else dataclasses.field(default=default,
                             metadata=dataclasses_json.config(**metadata))
  )


def metadata(base: Mapping[str, Any], **custom) -> Mapping[str, Any]:
  for key in custom:
    if not custom[key] and key in base:
      del base[key]

    elif custom[key]:
      base.update({key: custom[key]})

  return base


def standard_field(default: Any = None, default_factory: Any = None,
                   **kwargs) -> dataclasses.Field:
  base = metadata({
      'letter_case': dataclasses_json.LetterCase.CAMEL,
      'exclude': lambda x: not x
  }, **kwargs)

  return field(default=default, default_factory=default_factory, **base)


def enum_field(default: Any = None, **kwargs) -> dataclasses.Field:
  base = {'encoder': lambda x: x.name if x else None, **kwargs}

  return standard_field(default=default, **base)


def list_field(default_factory: Any = list,
               **kwargs) -> dataclasses.Field:
  def __value(f: Any) -> Any:
    for a in ['render', 'to_dict']:
      if (m := getattr(f, a, None)) and callable(m):
        return m()

  base = {'encoder': lambda x: [__value(f) or f for f in x], **kwargs}

  return standard_field(default_factory=default_factory, **base)


class AutoNumber(enum.Enum):
  def __repr__(self):
    return f'{self.__class__.__name__}, {self.name}'

  def __new__(cls, *args, **kwargs):
    value = len(cls.__members__) + 1
    obj = object.__new__(cls)
    obj._value_ = value
    return obj


class Renderable(object):
  """Renderable adds a 'render' method to subclasses objects.

  Subclasses can also define the following special values, which can be set
  at runtime by the user as well if need be (although I can't think why):
  __SUPPRESS_TAG__ (bool)
    This causes the render method to behave like `to_dict`.

  __TAG_OVERRIDE__ (str)
    Renames the root tag from the camelCase class name to the specified string.

  Thus, given a fragment like this:
  ```
  class SampleWidget(Renderable):
    sample_tag: str = standard_field()

  s = Sample(sample_tag='Hello, my name is Inigo Montoya.')
  s.render()
  ```
  you would get
  `{'sampleWidget': {'sampleTag': 'Hello, my name is Inigo Montoya.'}}`

  However if `SampleWidget` were defined as:
  ```
  class SampleWidget(Renderable):
    __SUPPRESS_TAG__ = True
    sample_tag: str = standard_field()
  ```
  you'd get
  `{'sampleTag': 'Hello, my name is Inigo Montoya.'}`

  If it had the override set, thus:
  ```
  class SampleWidget(Renderable):
    __TAG_OVERRIDE__ = 'aSampleWidgetClass'
    sample_tag: str = standard_field()
  ```
  the `render` command would produce
  `{'aSampleWidgetClass': {'sampleTag': 'Hello, my name is Inigo Montoya.'}}`

  NOTE: the __TAG_OVERRIDE is *NOT* camel-cased. What you enter is what you get.

  A subclass can implement their own `render` method, but it must return the
  valid Chat API JSON. An examnple of this is the `Card` class which has to add
  the `cardId` tag level with the `card` itself at the JSON top level.
  """

  def render(self) -> Mapping[str, Any]:
    """Renders the widget in a usable form.

    Returns:
        Mapping[str, Any]: the json representation of the widget
    """
    if getattr(self, '__SUPPRESS_TAG__', False):
      return self.to_dict()

    render = {
        (getattr(self, '__OVERRIDE_TAG__', False) or
          stringcase.camelcase(self.__class__.__name__)): self.to_dict()}
    properties = inspect.getmembers(self.__class__,
                                    lambda v: isinstance(v, property))
    for (name, value) in properties:
      if widget_value := value.fget(self):
        render[stringcase.camelcase(name)] = widget_value

    return render



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses

import dataclasses_json
from card_framework import AutoNumber, enum_field, standard_field

from ..widget import Widget
from .action import Action


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class SwitchControl(Widget):
  """SwitchControl
  """
  class ControlType(AutoNumber):
    """ControlType _summary_
_
    """
    SWITCH = ()
    CHECK_BOX = ()

    @classmethod
    def _missing_(cls, value) -> SwitchControl.ControlType:
      """Backward compatilbility for old enums.

      If the old product names are still in use, replace them with
      the new values seamlessly.

      Args:
          value (str): enum string value requested

      Returns:
          ControlType: the corrected type

      Raises:
          ValueError if it was simply an incorrect enum rather an old value
      """
      if value == 'CHECKBOX':
        # Deprecated in favor of CHECK_BOX.
        return cls.CHECK_BOX

  name: str = standard_field(default='')
  value: str = standard_field(default='')
  selected: bool = standard_field(default=None)
  onChangeAction: Action = standard_field(default=None)
  controlType: ControlType = enum_field()



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses

import dataclasses_json
from card_framework import AutoNumber, enum_field, standard_field


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class Space(object):
  class SpaceType(AutoNumber):
    SPACE_TYPE_UNSPECIFIED = ()
    SPACE = ()
    GROUP_CHAT = ()
    DIRECT_MESSAGE = ()

  name: str = standard_field()
  type: str = standard_field()  # deprecated
  spaceType: SpaceType = enum_field()
  singleUserBotDm: bool = standard_field()
  threaded: bool = standard_field()
  displayName: str = standard_field()
  spaceDetails: SpaceDetail = standard_field()


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class SpaceDetail(object):
  description: str = standard_field()
  guidelines: str = standard_field()



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from dataclasses import dataclass
from typing import List

from card_framework import AutoNumber, enum_field, list_field, standard_field
from dataclasses_json import dataclass_json

from ..widget import Widget


@dataclass_json
@dataclass
class Action(Widget):
  class Interaction(AutoNumber):
    UNSPECIFIED = ()
    OPEN_DIALOG = ()

  class LoadIndicator(AutoNumber):
    """LoadIndicator
    """
    SPINNER = ()
    NONE = ()

  function: str = standard_field(default='')
  parameters: List[ActionParameter] = list_field()
  load_indicator: LoadIndicator = enum_field()
  persist_values: bool = standard_field()
  interaction: Interaction = enum_field()


@dataclass_json
@dataclass
class ActionParameter(object):
  key: str = standard_field()
  value: str = standard_field()



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses

import dataclasses_json
from card_framework import Renderable, enum_field

from .enums import HorizontalAlignment


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class Widget(Renderable):
  """Widget
  """
  __horizontal_alignment: HorizontalAlignment = enum_field(
    exclude=lambda x: True)

  @property
  def horizontal_alignment(self) -> str:
    return \
        self.__horizontal_alignment.name if self.__horizontal_alignment else None

  @horizontal_alignment.setter
  def horizontal_alignment(self, a: HorizontalAlignment = None) -> None:
    self.__horizontal_alignment = a



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import unittest
from dataclasses import Field, dataclass
from typing import List

from dataclasses_json import dataclass_json

from card_framework import *
from card_framework.v2.action_response import ActionResponse
from card_framework.v2.action_status import ActionStatus


class LazyPropertyTest(unittest.TestCase):
  class Foo(object):
    @lazy_property
    def thing(self) -> str:
      return 'lazy'

  def test_lazy_thing(self):
    foo = LazyPropertyTest.Foo()
    self.assertFalse(hasattr(foo, 'lazy_thing'))
    self.assertEqual('lazy', foo.thing)
    self.assertTrue(hasattr(foo, '_lazy_thing'))
    with self.assertRaises(AttributeError):
      foo.thing = 'bar'


class TestEnum(enum.Enum):
  ONE = enum.auto()
  TWO = enum.auto()


class MetadataTest(unittest.TestCase):
  def test_simple_update(self) -> None:
    base = {'pirate': 'Westley'}

    updated = metadata(base=base, pirate='Dread Pirate Roberts')

    self.assertDictEqual(updated, {'pirate': 'Dread Pirate Roberts'})

  def test_remove(self) -> None:
    base = {'pirate': 'Dread Pirate Roberts', 'swordsman': 'Inigo Montoya'}

    updated = metadata(base=base, swordsman=None)

    self.assertDictEqual(updated, {'pirate': 'Dread Pirate Roberts'})

  def test_add(self) -> None:
    base = {'pirate': 'Dread Pirate Roberts', }

    updated = metadata(base=base, swordsman='Inigo Montoya')

    self.assertDictEqual(updated, {'pirate': 'Dread Pirate Roberts',
                                   'swordsman': 'Inigo Montoya'})

  def test_add_and_remove(self) -> None:
    base = {'pirate': 'Dread Pirate Roberts', 'strongman': 'Fezzik'}

    updated = metadata(base=base, swordsman='Inigo Montoya', strongman=None)

    self.assertDictEqual(updated, {'pirate': 'Dread Pirate Roberts',
                                   'swordsman': 'Inigo Montoya'})

  def test_add_update_and_remove(self) -> None:
    base = {'pirate': 'Westley', 'strongman': 'Fezzik'}

    updated = metadata(base=base, swordsman='Inigo Montoya', strongman=None,
                       pirate='Dread Pirate Roberts')

    self.assertDictEqual(updated, {'pirate': 'Dread Pirate Roberts',
                                   'swordsman': 'Inigo Montoya'})


class StandardFieldTest(unittest.TestCase):
  def test_base_field(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _field: str = standard_field()

    base = Base()
    base._field = 'foo'

    f: Field = base.__dataclass_fields__.get('_field')
    self.assertIsNone(f.default)
    self.assertTrue(isinstance(f.default_factory, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])

  def test_field_with_metadata_removed(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _field: str = standard_field(exclude=None)

    base = Base()
    base._field = 'foo'

    f = base.__dataclass_fields__.get('_field')
    self.assertIsNone(f.default)
    self.assertTrue(isinstance(f.default_factory, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertNotIn('exclude', f.metadata['dataclasses_json'])

  def test_field_with_metadata_edited(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _field: str = standard_field(exclude=True)

    base = Base()
    base._field = 'foo'

    f = base.__dataclass_fields__.get('_field')
    self.assertIsNone(f.default)
    self.assertTrue(isinstance(f.default_factory, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertTrue(f.metadata['dataclasses_json']['exclude'])

  def test_field_with_default(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _field: str = standard_field(default='Princess Buttercup')

    base = Base()

    f = base.__dataclass_fields__.get('_field')
    self.assertIsNotNone(f.default)
    self.assertTrue(isinstance(f.default_factory, dataclasses._MISSING_TYPE))
    self.assertEquals(base._field, 'Princess Buttercup')


class EnumFieldTest(unittest.TestCase):
  class Fencer(AutoNumber):
    DREAD_PIRATE_ROBERTS = ()
    INIGO_MONTOYA = ()
    SIX_FINGERED_MAN = ()

  def test_base_field(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _field: EnumFieldTest.Fencer = enum_field()

    base = Base()
    base._field = EnumFieldTest.Fencer.INIGO_MONTOYA

    f: Field = base.__dataclass_fields__.get('_field')
    self.assertIsNone(f.default)
    self.assertTrue(isinstance(f.default_factory, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])
    self.assertIn('encoder', f.metadata['dataclasses_json'])
    self.assertEquals(base._field, EnumFieldTest.Fencer.INIGO_MONTOYA)
    self.assertDictEqual(base.to_dict(), {'field': 'INIGO_MONTOYA'})

  def test_field_with_edited_encoder(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _enum_field: EnumFieldTest.Fencer = enum_field(
          encoder=lambda x: x.value if x else None)

    base = Base()
    base._enum_field = EnumFieldTest.Fencer.SIX_FINGERED_MAN

    f: Field = base.__dataclass_fields__.get('_enum_field')
    print(EnumFieldTest.Fencer.SIX_FINGERED_MAN.value)
    print(ActionStatus.Code.ABORTED.value)
    self.assertIsNone(f.default)
    self.assertTrue(isinstance(f.default_factory, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])
    self.assertIn('encoder', f.metadata['dataclasses_json'])
    self.assertDictEqual(
        base.to_dict(),
        {'enumField': EnumFieldTest.Fencer.SIX_FINGERED_MAN.value})

  def test_field_with_metadata_removed(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _enum_field: EnumFieldTest.Fencer = enum_field(letter_case=None)

    base = Base()
    base._enum_field = EnumFieldTest.Fencer.SIX_FINGERED_MAN

    f: Field = base.__dataclass_fields__.get('_enum_field')
    self.assertIsNotNone(f.metadata)
    self.assertNotIn('letter_case', f.metadata['dataclasses_json'])
    self.assertDictEqual(
        base.to_dict(),
        {'_enum_field': EnumFieldTest.Fencer.SIX_FINGERED_MAN.name})


class ListFieldTest(unittest.TestCase):
  def test_base_list_field_str(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _list_field: List[str] = list_field()

    base = Base()
    base._list_field = 'Hello, my name is Inigo Montoya'.split(' ')

    f: Field = base.__dataclass_fields__.get('_list_field')
    self.assertEquals(f.default_factory, list)
    self.assertTrue(isinstance(f.default, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])
    self.assertIn('encoder', f.metadata['dataclasses_json'])
    self.assertListEqual(
        'Hello, my name is Inigo Montoya'.split(' '), base._list_field)
    self.assertDictEqual(
        base.to_dict(),
        {'listField': ['Hello,', 'my', 'name', 'is', 'Inigo', 'Montoya']})

  def test_base_list_field_int(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _list_field: List[int] = list_field()

    base = Base()
    base._list_field = [1, 2, 3, 4, 5, ]

    f: Field = base.__dataclass_fields__.get('_list_field')
    self.assertEquals(f.default_factory, list)
    self.assertTrue(isinstance(f.default, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])
    self.assertIn('encoder', f.metadata['dataclasses_json'])
    self.assertListEqual(
        [1, 2, 3, 4, 5, ], base._list_field)
    self.assertDictEqual(
        base.to_dict(),
        {'listField': [1, 2, 3, 4, 5]})

  def test_base_list_field_no_render(self) -> None:
    @dataclass_json
    @dataclass
    class Base(object):
      _list_field: List[ActionStatus] = list_field()

    LIST_UNDER_TEST = [
        ActionStatus(status_code=ActionStatus.Code.OK,
                     user_facing_message='Hello, my name is Inigo Montoya'),
        ActionStatus(status_code=ActionStatus.Code.UNKNOWN,
                     user_facing_message='Inconcievable!'),
    ]
    base = Base()
    base._list_field = LIST_UNDER_TEST

    f: Field = base.__dataclass_fields__.get('_list_field')
    self.assertEquals(f.default_factory, list)
    self.assertTrue(isinstance(f.default, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])
    self.assertIn('encoder', f.metadata['dataclasses_json'])
    self.assertListEqual(
        LIST_UNDER_TEST, base._list_field)

    self.assertDictEqual(
        base.to_dict(),
        {'listField': [{'statusCode': 'OK',
                        'userFacingMessage': 'Hello, my name is Inigo Montoya'},
                       {'statusCode': 'UNKNOWN',
                        'userFacingMessage': 'Inconcievable!'}]})

  def test_base_list_field_render(self) -> None:
    """test_base_list_field_render

    ActionResponse has a render method, which causes the tag 'actionResponse' to
    be the root of each rendered element unlike a `to_dict` call which would
    drop the camel-cased class name.
    """
    @dataclass_json
    @dataclass
    class Base(object):
      _list_field: List[ActionResponse] = list_field()

    LIST_UNDER_TEST = [
        ActionResponse(
            type=ActionResponse.ResponseType.NEW_MESSAGE,
            url='http://www.karentaylorart.com'
        ),
        ActionResponse(
            type=ActionResponse.ResponseType.NEW_MESSAGE,
            url='http://www.imdb.com/title/tt0093779/'
        ),
    ]
    base = Base()
    base._list_field = LIST_UNDER_TEST

    f: Field = base.__dataclass_fields__.get('_list_field')
    self.assertEquals(f.default_factory, list)
    self.assertTrue(isinstance(f.default, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])
    self.assertIn('encoder', f.metadata['dataclasses_json'])
    self.assertListEqual(
        LIST_UNDER_TEST, base._list_field)

    self.assertDictEqual(
        base.to_dict(),
        {'listField': [
            {'actionResponse': {'type': 'NEW_MESSAGE',
                                'url': 'http://www.karentaylorart.com'}},
            {'actionResponse': {'type': 'NEW_MESSAGE',
                                'url': 'http://www.imdb.com/title/tt0093779/'}}
        ]})

  def test_list_with_thing_with_render_property(self) -> None:
    @dataclass_json
    @dataclass
    class Thing(object):
      render: str = standard_field()

    @dataclass_json
    @dataclass
    class Base(object):
      _list_field: List[Thing] = list_field()

    LIST_UNDER_TEST = [
        Thing(render='Florin'),
        Thing(render='Guilder')
    ]
    base = Base()
    base._list_field = LIST_UNDER_TEST

    print(Thing(render='Florin').to_dict())
    f: Field = base.__dataclass_fields__.get('_list_field')
    self.assertEquals(f.default_factory, list)
    self.assertTrue(isinstance(f.default, dataclasses._MISSING_TYPE))
    self.assertIsNotNone(f.metadata)
    self.assertIn('letter_case', f.metadata['dataclasses_json'])
    self.assertIn('exclude', f.metadata['dataclasses_json'])
    self.assertIn('encoder', f.metadata['dataclasses_json'])
    self.assertListEqual(
        LIST_UNDER_TEST, base._list_field)

    self.assertDictEqual(
        base.to_dict(),
        {'listField': [
            {'render': 'Florin'},
            {'render': 'Guilder'},
        ]})


class RenderableTest(unittest.TestCase):
  def test_render(self) -> None:
    @dataclass_json
    @dataclass
    class Base(Renderable):
      _field: str = standard_field()

    base = Base(_field='Hello, my name is Inigo Montoya.')

    self.assertDictEqual(
        base.render(),
        {'base': {'field': 'Hello, my name is Inigo Montoya.'}})

  def test_render_no_tag_name(self) -> None:
    @dataclass_json
    @dataclass
    class Base(Renderable):
      __SUPPRESS_TAG__ = True
      _field: str = standard_field()

    base = Base(_field='Hello, my name is Inigo Montoya.')

    self.assertDictEqual(
        base.render(),
        {'field': 'Hello, my name is Inigo Montoya.'})

  def test_render_tag_override(self) -> None:
    @dataclass_json
    @dataclass
    class Base(Renderable):
      __OVERRIDE_TAG__ = 'overriddenTagName'
      _field: str = standard_field()

    base = Base(_field='Hello, my name is Inigo Montoya.')

    self.assertDictEqual(
        base.render(),
        {'overriddenTagName': {'field': 'Hello, my name is Inigo Montoya.'}})



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import Optional

import dataclasses_json
from card_framework import standard_field

from .color import Color
from .icon import Icon
from .on_click import OnClick


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class Button(object):
  """Button

  Renders a Button.

  https://developers.google.com/chat/api/guides/message-formats/cards#buttons
  """
  text: str = standard_field()
  icon: Icon = standard_field()
  color: Optional[Color] = standard_field()
  on_click: OnClick = standard_field()
  disabled: bool = standard_field()
  alt_text: str = standard_field()



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import Optional

from card_framework import AutoNumber, enum_field, standard_field
import dataclasses_json

from ..widget import Widget
from .action import Action
from .suggestions import Suggestions


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class TextInput(Widget):
  """TextInput
  """
  class Type(AutoNumber):
    """TextInputType
    """
    SINGLE_LINE = ()
    MULTIPLE_LINE = ()

  name: str = standard_field()
  label: Optional[str] = standard_field()
  hint_text: Optional[str] = standard_field()
  value: Optional[str] = standard_field()
  type: Type = enum_field()
  on_change_action: Optional[Action] = standard_field()
  initial_suggestions: Optional[Suggestions] = standard_field()
  auto_complete_action: Optional[Action] = standard_field()



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import Optional

import dataclasses_json
from card_framework import standard_field

from ..widget import Widget
from .on_click import OnClick


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class Image(Widget):
  """Image widget.

  Renders an Image widget

  https://developers.google.com/chat/api/reference/rest/v1/cards#image
  """
  image_url: str = standard_field()
  on_click: Optional[OnClick] = standard_field()
  alt_text: str = standard_field()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from card_framework import AutoNumber

# enums.py
#
# Generic enums, reused in multiple classes. Most enums are widget specific,
# so are defined in that widget.


class ImageType(AutoNumber):
  """ImageType

  The possible image styles of a Header image.
  """
  SQUARE = ()
  CIRCLE = ()


class HorizontalAlignment(AutoNumber):
  """HorizontalAlignment
  """
  START = ()
  CENTER = ()
  END = ()

  def __str__(self) -> str:
    return self.name

  def __repr__(self) -> str:
    return str(self.name)



# Copyright 2022 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import uuid
from typing import List, Optional

import dataclasses_json
from card_framework import AutoNumber, Renderable, list_field, standard_field

from .card_action import CardAction
from .card_fixed_footer import CardFixedFooter
from .card_header import CardHeader
from .section import Section


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class RenderableCard(Renderable):
  __card_id = None

  @property
  def card_id(self) -> str:
    return self.__card_id if self.__card_id else str(uuid.uuid4())

  @card_id.setter
  def card_id(self, value: str) -> None:
    self.__card_id = value

  card: Card = standard_field(exclude=lambda x: True)


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class Card(RenderableCard):
  """Response

  A response object that can be `render`ed to produce a valid Google Chat App
  json response.

  See https://developers.google.com/chat/api/guides/message-formats/cards for
  full details on what this should look like.
  """
  class DisplayStyle(AutoNumber):
    DISPLAY_STYLE_UNSPECIFIED = ()
    PEEK = ()
    REPLACE = ()

  header: Optional[CardHeader] = standard_field()
  name: Optional[str] = standard_field()
  sections: Optional[List[Section]] = list_field(default_factory=list)
  card_actions: Optional[List[CardAction]] = standard_field()
  fixed_footer: Optional[CardFixedFooter] = standard_field()
  display_style: Optional[DisplayStyle] = standard_field()
  peek_card_header: Optional[CardHeader] = standard_field()

  def add_section(self, section: Section) -> None:
    """Adds a section to the report.

    Args:
        section (Section): The section to add.
    """
    self.sections.append(section)
